#!/usr/bin/perl

use strict;
use warnings;

package VProtect::Backup;
use Data::Dumper;

use VProtect qw/vim_get_virtual_machines log_msg/;
use VMware::Vix::Simple;
use VMware::Vix::API::Constants;

use POSIX "strftime";

my $gos_guestPath;
my $gos_guestVssPgmPath;
my $gos_vmHandle;
my $gos_hostHandle;
my $gos_osLevel;


sub get_moref_by_uuid ($$$) {
	my $vmuuid = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;

	my $vm_moref = VProtect::vim_get_vm_by_uuid($vmuuid);

	if (!defined $vm_moref) {
		die "VM with UUID " . $vmuuid . " not found.\n" .
		"Try refreshing the list of Virtual Machines\n";
	}

	my $vm_xmlelem = $xmldoc->createElement('vm_info');
	$vm_xmlelem->appendTextChild('vm_moref', $vm_moref->{'value'});
	$vm_xmlelem->appendTextChild('vm_moref_type', $vm_moref->{'type'});
	$xmlroot->addChild($vm_xmlelem);

	return $vm_moref;
}

sub get_template_moref_by_uuid ($$$) {
	my $template_uuid = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;

	my $template_moref = VProtect::vim_get_template_by_uuid($template_uuid);

	if (!defined $template_moref) {
		die "Template with UUID " . $template_uuid . " not found.\n" .
		"Try refreshing the list of Virtual Machines\n";
	}

	# Attach similar elements in the XML as those for a VM,
	# so that the complete parsing need not be changed
	my $vm_xmlelem = $xmldoc->createElement('vm_info');
	$vm_xmlelem->appendTextChild('vm_moref', $template_moref->{'value'});
	$vm_xmlelem->appendTextChild('vm_moref_type', $template_moref->{'type'});
	$xmlroot->addChild($vm_xmlelem);

	return $template_moref;
}

sub purge_snaps ($$$);  # required to get type checking
sub purge_snaps ($$$) {
	# Remove snapshots named $snapshot_name
	my ($snapshot_tree, $ch_present_ref, $snapshot_name) = @_;

	if ($snapshot_tree->childSnapshotList) {
		foreach (@{$snapshot_tree->childSnapshotList}) {
			purge_snaps($_, $ch_present_ref, $snapshot_name);
		}
	}

	if ($snapshot_tree->name eq $snapshot_name) {
		my $snap_view = Vim::get_view(mo_ref => $snapshot_tree->snapshot);
		eval {
			$snap_view->RemoveSnapshot(removeChildren => 0);
		}; if ($@) {
			# This is a fatal error.
			my $hdr = "Could not remove previous Unitrends snapshot";

			if (ref($@->detail) eq 'FileLocked') {
				die "$hdr: the file is locked: " . Dumper($@) . "\n";
			}
			elsif (ref($@->detail) eq 'ManagedObjectNotFound') {
				# This should probably not be a fatal error, but we have to be conservative with snapshot removal issues.
				die "$hdr: The object has already been deleted or has not been completely created.\n";
			}
			elsif (ref($@->detail) eq 'NoPermission') {
				die "$hdr: permission denied.  Requires permission " . $@->detail->privilegeId . "\n";
			}
			elsif (ref($@->detail) eq 'SystemError') {
				die "$hdr: " . $@->fault_string . "\n";
			}
			elsif (ref($@->detail) eq 'SystemError') {
				die "Could not remove previous Unitrends snapshot: A general system error occurred: " . $@->detail->reason . "\n";
			}
			else {
				die "$hdr: " . Dumper($@) . "\n";
			}
		}
	}
	elsif ($snapshot_tree->name =~ "Consolidate Helper.*") {
		# Flag that consolidate helper snapshot is present.	
		$$ch_present_ref = 1;
	}
}

# This sub-routine is similar to 'configure_change_tracking' sub-routine
# However, this will be called from 'validate_hw_hash' only for Incr/Diff Backups
sub reset_cbt ($$) {

	my $vm_moref = shift;
	my $xmlroot = shift;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $has_physical_rdm_disks = physical_rdm_disks_present($vm_view);

	if (not $vm_view->capability->changeTrackingSupported) {
		disable_change_tracking($vm_view);
		print STDERR "This VM does not support change tracking.\n";
		return;
	} else {
		# Change Tracking is supported by the ESX on which this VM is
		if ($vm_view->config->changeTrackingEnabled) { # Safety check
			# Change Tracking is currently enabled on this VM
			print STDERR "  Temporarily disabling change tracking.\n";
			disable_change_tracking($vm_view);

			print STDERR "  Using snapshot create/delete cycle to stun VM.\n";
			snapshot_stun_cycle($vm_view);

			print STDERR "  Re-enabling change tracking.\n";
			enable_change_tracking($vm_view, $xmlroot);
		} else {
			print STDERR "CBT is currently disabled on this VM.\n";
		}
	}

}

sub validate_hw_hash ($$$$$$) {
	my $prev_hw_hash = shift;
	my $new_hw_hash = shift;
	my $is_full = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;
	my $vm_moref = shift;

	if ($is_full) {
		# Don't check
		return;
	}

	if ((not defined $prev_hw_hash) or ($prev_hw_hash ne $new_hw_hash)) {
		# Signal that next backup should be promoted to full
		VProtect::promote_next_reason($xmldoc, 'A VM configuration change has been detected.');

		# Reset CBT before dying
		print STDERR "Resetting CBT since a configuration change has been detected.\n";
		reset_cbt($vm_moref, $xmlroot);

		# Fail backup
		die "VM configuration has changed, a full backup is required.\n";
	}
}

sub check_for_fault_tolerance ($) {
	# Fail the backup if the VM is configured for Fault Tolerance.
	# This is unsupported because snapshots are disabled.

	my $vm_moref = shift;
	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	if (defined $vm_view->config->ftInfo) {
		die "This VM cannot be backed up because it is configured for Fault Tolerance.\n";
	}
}

sub clean_up_snapshots ($$$$) {
	# Remove old 'Unitrends' snapshots left from a previous backup that was terminated uncleanly.
	my ($vm_moref, $xmldoc, $xmlroot, $warning_ref) = @_;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $snap_view = $vm_view->snapshot;

	my $ch_present = 0;  # Consolidate helper snapshot is present
	my $snapshot_name = "Unitrends_backup";

	if (defined $vm_view->snapshot) {
		if (defined (@{$vm_view->snapshot->rootSnapshotList})) {
			foreach (@{$vm_view->snapshot->rootSnapshotList}) {
				purge_snaps($_, \$ch_present, $snapshot_name);
			}
		}
	}

	if ($ch_present) {
		my $ch_xmlelem = $xmldoc->createElement('consolidate_helper_info');
		$ch_xmlelem->appendTextChild('consolidate_helper_snapshot_present', 1);
		$xmlroot->addChild($ch_xmlelem);
		$$warning_ref .= "Warning: A consolidate helper snapshot is present on the VM.  Presence of this snapshot indicates a possible issue with the snapshot system for this VM.  Please remove this snapshot to ensure that backups function properly and do not affect the VM.\n\n";
	}
}

sub enable_change_tracking ($$) {
	my $vm_view = shift;
	my $xmlroot = shift;
	
	if ($vm_view->capability->changeTrackingSupported) {
		my $spec = VirtualMachineConfigSpec->new(changeTrackingEnabled => 'true');
		my $task_ref;

		eval {
			# Use a task since ReconfigVM() doesn't wait for completion.
			$task_ref = $vm_view->ReconfigVM_Task(spec => $spec);
		};
		if ($@) {
			if (ref($@) eq 'SoapFault') {
				if (ref($@->detail) eq 'RestrictedVersion') {
					die "Error enabling change tracking on VM: RestrictedVersionFault\n" .
					    "This host is not licensed to perform this operation.\n";
				} elsif (ref($@->detail) eq 'NoPermission') {
					die "Error enabling change tracking on VM: permission to perform this operation was denied.\n" .
					$@->detail->privilegeId . " permission is required\n";
				}
			}

			die "Error enabling change tracking on VM:" . $@ . "\n";
		}

		my $task = Vim::get_view(mo_ref => $task_ref);

		sleep 2;

		while (($task->info->state->val eq 'running') or
				($task->info->state->val eq 'queued')) {
			sleep 2;
			$task->update_view_data();
		}

		if ($task->info->state->val eq 'success') {
			# change tracking was enabled
			$xmlroot->appendTextChild('currentChangeTrackingEnabled', '1');
			print STDERR "Change tracking was enabled\n";
		} elsif ($task->info->state->val eq 'error') {
			print STDERR "Could not enable change tracking.\n";
			print STDERR Dumper($task);
			my $error_msg = "<no further information provided>";
			if (defined($task->info->error->localizedMessage)) {
				$error_msg = $task->info->error->localizedMessage;
			}
			die "Could not enable change tracking: $error_msg\n";
		} else {
			die "Unknown state during ReconfigVM task: State->val=" . $task->info->state->val . "\n";
		}
	}
}

sub disable_change_tracking ($) {
	my $vm_view = shift;

	# bug 18622 always turn off changeTrackingEnabled regardless of changeTrackingSuppported.
#	if ($vm_view->capability->changeTrackingSupported) {
		if ($vm_view->config->changeTrackingEnabled) {
			my $spec = VirtualMachineConfigSpec->new(changeTrackingEnabled => 'false');
			# Use a task since ReconfigVM() doesn't wait for completion.
			my $task_ref = $vm_view->ReconfigVM_Task(spec => $spec);
			my $task = Vim::get_view(mo_ref => $task_ref);

			sleep 2;

			while (($task->info->state->val eq 'running') or 
					($task->info->state->val eq 'queued')) {
				sleep 2;
				$task->update_view_data();
			}

			if ($task->info->state->val eq 'success') {
				# change tracking was disabled
				print STDERR "Change tracking was disabled\n";
			} elsif ($task->info->state->val eq 'error') {
				my $error_msg = "<no further information provided>";
				if (defined($task->info->error->localizedMessage)) {
					$error_msg = $task->info->error->localizedMessage;
				}
				print STDERR "Could not disable change tracking: $error_msg.\n";
				print STDERR Dumper($task);
				# Do not abort.
			} else {
				die "Disable cbt: State->val=" . $task->info->state->val . "\n";
			}
		}
#	}
}

sub snapshot_stun_cycle ($) {
	# Create a snapshot and remove it to "stun" the VM.
	# Do not quiesce since we aren't using this snapshot for backup.

	my $vm_view = shift;

	my $snap_moref;

	eval {
	$snap_moref = $vm_view->CreateSnapshot(
			name => "Unitrends_backup",
			description => "Created for backup to Unitrends DPU. (temporary snapshot)",
			memory => 0,
			quiesce => 0 );
	};
	if ($@) {
		if (ref($@) eq 'SoapFault') {
			print STDERR "Error creating temporary snapshot: " . $@ . "\n";
		}

		return;
	}

	eval {
		my $snap_view = Vim::get_view(mo_ref => $snap_moref);
		$snap_view->RemoveSnapshot(removeChildren => 0);
	};
	if ($@) {
		if (ref($@) eq 'SoapFault') {
			print STDERR "Error removing temporary snapshot: " . $@ . "\n";
			# TODO: should parse faults like other RemoveSnapshot calls.
		}
	}
}

sub snapshots_exist ($) {
	# Return 0 if no snapshots exist, 1 if some do.

	my $vm_view = shift;

	if (defined $vm_view->snapshot) {
		print STDERR "Snapshots do exist on this VM.\n";
		return 1;
	} else {
		print STDERR "Snapshots do not currently exist on this VM.\n";
		return 0;
	}
}

sub independent_disks_present ($) {

	my $vm_view = shift;

	my $devices = $vm_view->config->hardware->device;

	foreach (@$devices) {
		if ($_->isa('VirtualDisk')) {
			if( $_->{'backing'}->diskMode ne 'persistent') {
				# Independent Disks have the value 'independent_persistent'
				# or 'independent_nonpersistent'
				# NOTE: Physical RDMs are independent disks too
				print STDERR "Located an Independent Disk.\n";
				return 1;
			}
		}
	}

	return 0;

}

sub physical_rdm_disks_present ($) {
	# Only use this call during backup, it's somewhat expensive.
	my $vm_view = shift;

	my $devices = $vm_view->config->hardware->device;

	foreach (@$devices) {
		if ($_->isa('VirtualDisk')) {
			if ($_->{'backing'}->isa('VirtualDiskRawDiskMappingVer1BackingInfo')) {
				if ( ($_->backing->{'compatibilityMode'}) eq "physicalMode") {
					print STDERR "Located a physical RDM disk.\n";
					return 1;
				}
			}
		}
	}

	return 0;
}

sub device_is_on_nfs ($) {
	my $device = shift;

	my $datastore = $device->backing->datastore;
	my $ds_view = Vim::get_view(mo_ref => $datastore);
	my $ds_info = $ds_view->{'info'};
	if (defined $ds_info) {
		if ($ds_info->can('nas')) {
			if ($ds_info->nas->type eq 'NFS') {
				print STDERR "Device " . $device->key . " is on NFS.\n";
				return 1;
			}
		}
	}

	if (defined ($device->backing->parent)) {
		my $pds_view = Vim::get_view(mo_ref => $device->backing->parent->datastore);
		my $pds_info = $pds_view->{'info'};
		if (defined $pds_info) {
			if ($pds_info->can('nas')) {
				if ($pds_info->nas->type eq 'NFS') {
					print STDERR "Device " . $device->key . " is on NFS.\n";
					return 1;
				}
			}
		}
	}

	return 0;
}

sub nfs_disks_present ($) {
	my $vm_view = shift;

	my $devices = $vm_view->config->hardware->device;

	foreach (@$devices) {
		if ($_->isa('VirtualDisk')) {
			if (device_is_on_nfs($_)) {
				print STDERR "Located an NFS disk.\n";
				return 1;
			}
		}
	}

	return 0;
}

sub configure_change_tracking ($$$$$$) {
	# enable change tracking if it is disabled

	my $vm_moref = shift;
	my $is_full = shift;
	my $xmlroot = shift;
	my $warning_ref = shift;
	my $retry = shift;
	my $promote_next_backup_because_cbt_off_ref = shift;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $has_physical_rdm_disks = physical_rdm_disks_present($vm_view);
	my $has_nfs_disks = nfs_disks_present($vm_view);

	if (not $vm_view->capability->changeTrackingSupported) {
		# Bug 18622 make sure changeTrackingEnabled is
		# off when changeTrackingSupported is false
		disable_change_tracking($vm_view);
		print STDERR "This VM does not support change tracking.\n";
		return 0;
	}

	if ($vm_view->capability->changeTrackingSupported) {
		if ($vm_view->config->changeTrackingEnabled) {
			print STDERR "Change Block Tracking (CBT) is already ENABLED on this VM.\n";
			if (snapshots_exist($vm_view) == 0) {
				# No snapshots currently exist on this VM

				if ($is_full) {

					if ($retry) {
						print STDERR "(RETRYING) Performing a FULL Backup of a VM with NO Snapshots.\n" .
						"Was not able to successfully get CBT information.\n" .
						"Now, performing stun-unstun cycle.\n";

						# On a "RETRY", disable and re-enable if we're doing a full backup
						# and there are no snapshots.
						# This is meant to clean up the issue where change tracking was
						# enabled while snapshots already existed.

						# Do not touch the change tracking settings if RDM disks are present.

						print STDERR "Temporarily disabling change tracking.\n";
						disable_change_tracking($vm_view);

						print STDERR "Using snapshot create/delete cycle to stun VM.\n";
						snapshot_stun_cycle($vm_view);

						print STDERR "Re-enabling change tracking.\n";
						enable_change_tracking($vm_view, $xmlroot);

						#print STDERR "Using snapshot create/delete cycle to stun VM.\n";
						#snapshot_stun_cycle($vm_view);	# This should not be required

					} else {
						print STDERR "Performing a FULL Backup of a VM with NO Snapshots.\n" .
						"Not performing the stun-unstun cycle.\n";
					}

				} else {
					print STDERR "Performing a Full Backup? ===> $is_full\n";
					print STDERR "VM has one or more pRDMS? ===> $has_physical_rdm_disks\n";
				}

			} else {
				print STDERR "Not reconfiguring Change Block Tracking (CBT) because snapshots exist.\n";
			}

			return 1;
		} else {
			print STDERR "Change Block Tracking (CBT) is currently DISABLED on this VM.\n";

			if (!$is_full) {
				# CBT is currently disabled on the VM and the request is to perform
				# an Incremental/Differential Backup. Can't do it.
				# FAIL current Backup and promote the next one to a Full.

				$$promote_next_backup_because_cbt_off_ref = 1;
				return;
			}

			# Performing a Full Backup
			if (snapshots_exist($vm_view) == 1) {
				# Snapshot/s exist/s currently on this VM
				# Enabling Change Block Tracking (CBT) while snapshots exist can cause problems.

				# perform a non-sparse full backup
				$$warning_ref = $$warning_ref .
					"Warning: Not enabling change tracking because snapshots exist.\n" .
					"Backups will contain all blocks on the VM's disk(s).\n"           . 
					"To enable optimized (sparse) backups, remove all snapshots\n"     .
					"from the VM and then perform a Full backup.\n"                    .
					"Snapshots may be created after this has been done.\n";

				print STDERR "Not enabling change tracking because snapshots exist.\n" .
				"Backup will contain all the blocks and will be the size of the disk/s.\n"

			} else {
				# No snapshots exist, enable change tracking.
				enable_change_tracking($vm_view, $xmlroot);
			}

			return 0;
		}
	} else {
		print STDERR "Change Block Tracking (CBT) is not supported on this ESX server.\n";
	}
}

sub handle_snapshot_error {
	# Die with message for errors that can occur with CreateSnapshot()
	# If $may_retry is 1, return 1 for errors that indicate that
	# a second attempt with memory snapshotting enabled, or quiescing
	# disabled, may succeed.

	my $may_retry = shift;  # 0 = always die, 1 = return 1 if retry is possible
	my (@error) = shift;    # Pass $@ by ref.

	my $err_hdr = "Error creating snapshot of VM";

	if (ref($@) eq 'SoapFault') {
		if (ref($@->detail) eq 'RestrictedVersion') {
			die "$err_hdr: This host is not licensed to perform this operation.\n";
		}
		if (ref($@->detail) eq 'ApplicationQuiesceFault') {
			if ($may_retry) {
				# Indicate caller should retry w/o quiescing
				return 1;
			} else {
				die "$err_hdr: Application quiescing failed.\n";
			}
		}
		if (ref($@->detail) eq 'FilesystemQuiesceFault') {
			if ($may_retry) {
				# Indicate caller should retry w/o quiescing
				return 1;
			} else {
				die "$err_hdr: Filesystem quiescing failed.\n";
			}
		}
		# PASSTHROUGH is added to print the '=' correctly in the KB link
		if (ref($@->detail) eq 'FileTooLarge') {
			die "$err_hdr:\n" .
				"File is larger than the maximum size supported by datastore.\n" .
				"PASSTHROUGHPlease see http://support.unitrends.com/ikm/questions.php?questionid=22 for possible steps to resolve this problem.\n";
		}
		if (ref($@->detail) eq 'GenericVmConfigFault') {
			if ($may_retry) {
				# Indicate caller should retry w/o quiescing
				return 1;
			} else {
				die "$err_hdr: GenericVmConfigFault: " . $@->detail->reason . "\n" .
				"PASSTHROUGHPlease see http://support.unitrends.com/ikm/questions.php?questionid=1449 for possible steps to resolve this problem.\n";
			}
		}
		if (ref($@->detail) eq 'HostCommunication') {
			die "$err_hdr: An error occurred while communicating with the remote host.\n";
		}
		if (ref($@->detail) eq 'HostNotConnected') {
			die "$err_hdr: Unable to communicate with remote host.\n";
		}
		if (ref($@->detail) eq 'InvalidName') {
			die "$err_hdr: The snapshot name is invalid.\n";
		}
		if (ref($@->detail) eq 'InvalidDiskFormat') {
			die "$err_hdr: The disk format is either corrupted or not supported.\n";
		}
		if (ref($@->detail) eq 'InvalidState') {
			die "$err_hdr: The operation is not allowed in the current state.\n";
		}
		if (ref($@->detail) eq 'MemorySnapshotOnIndependentDisk') {
			die "$err_hdr: Memory snapshot requested for VM with an independent disk.\n";
		}
		if (ref($@->detail) eq 'MultipleSnapshotsNotSupported') {
			die "$err_hdr: This VM only supports a single snapshot at a time.\n";
		}
		if (ref($@->detail) eq 'NoPermission') {
			die "$err_hdr: Permission to perform this operation was denied: " . $@->detail->privilegeId . " permission required.\n";
		}
		if (ref($@->detail) eq 'NotSupported') {
			die "$err_hdr: Host does not support snapshots.\n";
		}
		if (ref($@->detail) eq 'SnapshotDisabled') {
			die "$err_hdr: Snapshots are disabled on the virtual machine.\n";
		}
		if (ref($@->detail) eq 'SnapshotIncompatibleDeviceInVm') {
			die "$err_hdr: An incompatible device exists on the virtual machine:\n" . $@->fault_string . "\n";
		}
		if (ref($@->detail) eq 'SnapshotLocked') {
			die "$err_hdr: The virtual machine's snapshot is locked.\n";
		}
		if (ref($@->detail) eq 'SnapshotNoChange') {
			die "$err_hdr: The virtual machine's state has not changed since the previous snapshot was created.\n";
		}
		if (ref($@->detail) eq 'SystemError') {
			# Often "Protocol Error from VMX"

			if ($may_retry) {
				return 1;
			} else {
				die "$err_hdr: System Error.\n" . $@;
			}
		}
		if (ref($@->detail) eq 'TaskInProgress') {
			if ($may_retry) {
				# This fault is sometimes returned when VSS has problems -- should retry.

				# Sleep for a short period of time in case this was a legitimate fault
				# caused by another task running.
				sleep 30;

				return 1;
			} else {
				die "$err_hdr: A task is already in progress.\n";
			}
		}
		if (ref($@->detail) eq 'Timedout') {
			if ($may_retry) {
				# Indicate to caller that operation should be retried.
				return 1;
			} else {
				die "$err_hdr: Operation timed out.\n";
			}
		}
		if (ref($@->detail) eq 'TooManySnapshotLevels') {
			die "$err_hdr: Too many snapshots exist.\n";
		}

		# Generic fault containers below, most faults should be caught by something above.

		if (ref($@->detail) eq 'SnapshotFault') {
			# Should be extended by one of the above faults
			die "$err_hdr: A snapshot operation cannot be performed. (SnapshotFault)\n";
		}

		if (ref($@->detail) eq 'VmConfigFault') {
			die "$err_hdr: VmConfigFault\n" . $@ . "\n";
		}

		if (ref($@->detail) eq 'FileAlreadyExists') {
			die "$err_hdr: FileAlreadyExists Fault.\n" .
			"PASSTHROUGHPlease see http://support.unitrends.com/ikm/questions.php?questionid=1439 " .
			"for possible steps to resolve this problem.\n" . $@ . "\n";
		}
	}

	# Failure was not caught above.
	die "$err_hdr: " . $@ .  "\n";
}


sub create_snapshot ($$$$$$) {
	# returns: moref of snapshot
	#  or dies with a failure message
	my $vm_moref = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;
	my $warning_ref = shift;
	my $err = shift;        # Error returned from do_guest_vss_snapshot
	my $has_independent_disks = shift;

	print STDERR "Value of err is $err.\n";
	print STDERR "Independent Disk/s present? => $has_independent_disks.\n";

	# Need to call CreateSnapshot on a view, not a moref.
	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $snap_moref;  # New moref of created snapshot.

	my $snap_name = "Unitrends_backup";
	my $timestamp = strftime("%H:%M %Z %a %b %e %Y", localtime);
	my $description = "Created for backup to Unitrends DPU: " . $timestamp;

	if ($err == 0) {
		# $err will be set to zero when appaware processing was successful
		print STDERR "Guest snapshot operation was successful. Will call CreateSnapshot with quiesce=0.\n";
		goto NOQUIESCE;
	}

	eval {
		# Pass quiesce = 1 so that vmware-tools can do
		# its VSS magic if installed.  If it is not installed,
		# things still work fine.

		print STDERR "Calling CreateSnapshot with memory=0, quiesce=1\n";

		$snap_moref = $vm_view->CreateSnapshot(
			name => $snap_name,
			description => $description,
			memory => 0,
			quiesce => 1 );
	};
	if ($@) {
		# Snapshot failed, process error.
		if (handle_snapshot_error(1, \$@) == 1) {

			if ($has_independent_disks) {
				print STDERR "CreateSnapshot operation with memory=0,quiesce=1 failed.\n" .
				"This VM has at least one Independent Disk. So, can't snapshot it with " .
				"memory=1. Will now attempt to CreateSnapshot with memory=0,quiesce=0.\n";
				goto NOQUIESCE;
			}

			# Error was related to quiescing, first, try again with memory = 1.
			# This works around some VSS quiescing problems.
			eval {
				print STDERR "Calling CreateSnapshot with memory=1, quiesce=1\n";

				$snap_moref = $vm_view->CreateSnapshot(
						name => $snap_name,
						description => $description,
						memory => 1,
						quiesce => 1 );
			};
			if ($@) {
				if (handle_snapshot_error(1, \$@) == 1) {
					# Error was related to quiescing, try again w/o quiescing,
					# and issue a warning.
					NOQUIESCE:
					eval {
						print STDERR "Calling CreateSnapshot with memory=0, quiesce=0\n";

						$snap_moref = $vm_view->CreateSnapshot(
								name => $snap_name,
								description => $description,
								memory => 0,
								quiesce => 0 );

						if ($err != 0) {
							# Add a warning message only when app-aware was not performed
							$$warning_ref = $$warning_ref . "\n" .
								"This backup encountered a snapshot timeout and was performed with a snapshot without I/O quiescing.\n";
						}
					};

					if ($@) {
						# Snapshot failed, process error and terminate.
						handle_snapshot_error(0, \$@);
					}
				}
			}
		}
	}

	return $snap_moref;
}

sub transport_mode {

	my $xmlroot = shift;

	my @contents;

	if (open(my $CONFIG_FILE, '<', '/usr/bp/bin/vprotect/mode.conf')) {

		while(<$CONFIG_FILE>) {
			chomp;
			@contents = split('=', $_);
			if (defined $contents[1]) {
				if ($contents[1] eq "true") {
					print STDERR "Using best Transport Mode for Backup\n";
					$xmlroot->appendTextChild('bestTransportMode', 'true');
				} elsif ($contents[1] eq "false") {
					print STDERR "Using NBD Transport Mode for Backup\n";
					$xmlroot->appendTextChild('bestTransportMode', 'false');
				} else {
					print STDERR "Unable to determine Transport Mode for Backup\n";
					print STDERR Dumper($contents[1]);
					die "Unable to determine transport mode for Backup from configuration file\n";
				}
			}
		}

		close($CONFIG_FILE)
			or warn "Unable to close configuration file: $!";

	} else {
		# Unable to open the mode.conf file for some reason
		# Log the message, use the default/best transport mode, but don't fail the Backups
		print STDERR "Unable to open the mode.conf file - $!\n";
		print STDERR "Using the default/best transport mode for Backup\n";
		$xmlroot->appendTextChild('bestTransportMode', 'true');
	}

}

##
## Kas 9/22/2012 : Bug 14809
## This method will search for the Unitrends_backup
## snapshot in the childSnapshotList moref.
##
sub find_unitrends_snap ($); 
sub find_unitrends_snap ($) {
	my $snap_tree = shift;
	my $snapRet;
	
	print STDERR "In find_unitrends_snap: " . ref($snap_tree) . " \n ";
	if ($snap_tree->childSnapshotList) {
		print STDERR "snap_tree defined\n";
		foreach (@{$snap_tree->childSnapshotList}) {
			print STDERR "snap_tree entry\n";
			$snapRet = find_unitrends_snap($_);
		}
	}

	if ($snap_tree->name eq "Unitrends_backup") {
		print STDERR "Found unitrends_snap\n";
		$snapRet = Vim::get_view(mo_ref => $snap_tree->snapshot);
	} else {
		print STDERR "Found leaf, not unitrends snap\n";
	}

	return $snapRet;
}

sub get_disk_info ($$$$$$$$$$) {
	# params: moref for VM, xml doc, xml root node
	my $vm_moref = shift;
	my $snap_moref = shift;
	my $change_id_hash_ref = shift;
	my $keys_array = shift;
	my $is_template = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;
	my $is_full = shift;
	my $promote_next_ref = shift;
	my $snapshots_exist = shift;

	print STDERR "Entering get_disk_info\n";

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my @disks;
	my $snap;
	my $devices;
	my $disks_backup_size = 0;

	my $reset_cbt;

	my @keys_array_ref = @$keys_array;
	
	print STDERR "keys_array is @keys_array_ref\n";

	my $disks_xmlelem = $xmldoc->createElement('disks');

	if ($is_template eq "NO") {
		##
		## Kas 9/22/2012 : Bug 14809
		## Typically the snapshot->currentSnapshot identifies the
		## newest snapshot.  If that entry is empty, then look for
		## the Unitreds_backup snapshot in the snapshot list.
		## currentSnapshot has been seen empty on ESXi 5.1.
		##
		my $snap_moref_t = $vm_view->snapshot->currentSnapshot;
		if (!defined $snap_moref_t) {
			print STDERR "get_disk_info: No currentSnapShot reference. Will search for UT snapshot.\n";
			if (defined (@{$vm_view->snapshot->rootSnapshotList})) {
				foreach (@{$vm_view->snapshot->rootSnapshotList}) {
					$snap = find_unitrends_snap( $_ );
					if (defined($snap)) {
						last;
					}
				}
			}
		} else {
			print STDERR "get_disk_info: Using currentSnapshot reference.\n";
			$snap = Vim::get_view(mo_ref => $snap_moref_t);
		}
		$devices = $snap->config->hardware->device;
	} else {
		$devices = $vm_view->config->hardware->device;
	}

	my $flag = 0;
	foreach (@$devices) {
		if ($_->isa('VirtualDisk')) {
			$flag = 0;
			my $current_disk_backup_size = 0;
			# Bug 16541 - Exclude a pRDM, when 'empty' exclude argument is received to vprotect
			my $backing_comp_mode = $_->backing->{'compatibilityMode'};
			if (defined $backing_comp_mode) {
				if ($backing_comp_mode eq 'physicalMode') {
					foreach my $iterator (@keys_array_ref) {
						if ($iterator eq 'empty') {
							print STDERR "Exclude argument is - $iterator\n";
							print STDERR "A VM with Physical RDM found and no arguments were passed to exclude. " .
							"Skipping this disk\n";
							$flag = 1;
						}
					}
				}
			}

			if ($flag == 1) {
				next;
			}

			my $disk = {};
			my $disk_xmlelem = $xmldoc->createElement('disk');

			$disk->{'key'} = $_->key;

			print STDERR "Dealing with disk with key " . $_->{'key'} . "\n";

			$disk_xmlelem->appendTextChild('key', $_->{'key'});
			$disk_xmlelem->appendTextChild('capacityInKB', $_->{'capacityInKB'});
			$disk_xmlelem->appendTextChild('fileName', $_->backing->{'fileName'});
			$disk_xmlelem->appendTextChild('label', $_->{'deviceInfo'}{'label'});

			# Excluding Disks. Won't get the changed block areas for these disks
			foreach my $iterator (@keys_array_ref) {
				if (($_->{'key'} eq $iterator) || ($_->{'backing'}->diskMode ne 'persistent')) {
					if ($iterator == 'empty') {
						print STDERR "Skipping an independent disk by default\n";
					} else {
						print STDERR "Skipping disk with key $iterator\n";
					}
					print STDERR "Skipping disk with key $iterator\n";
					$flag = 1;
					last;
				}
			}
			
			if ($flag == 1)	{
				next;
			}

			###
			# Get the changed disk areas	
			###
			eval {
				($reset_cbt, $current_disk_backup_size) = get_changed_blocks_for_disk($vm_moref, $snap_moref, $_->{'key'}, $change_id_hash_ref->{$_->{'key'}}, $is_template, $xmldoc, $disk_xmlelem, $is_full, $promote_next_ref, $_->{'capacityInKB'}, $snapshots_exist);

				if ($reset_cbt == 33) {
					print STDERR "\tResetting CBT since got an error related to CBT.\n";
					$xmlroot->removeChild($disks_xmlelem);
				}

			};
			if ($@) {
				die "Failed to get changed blocks for disk. Remove any snapshots that may exist for the guest: " . $@ . "\n";
			}

			if ($reset_cbt == 33) {
				print STDERR "\tWill perform a retry to get CBT information.\n";
				last;	# Exit 'foreach (@$devices)', since have to retrieve CBT
					# information _again_ for ALL disks
			} else {
				print STDERR "Successfully retrieved CBT information.\n";
				push @disks, $disk;
				$disks_xmlelem->addChild($disk_xmlelem);
				$disks_backup_size = $disks_backup_size + $current_disk_backup_size;
			}
		}
	}

	if ($reset_cbt == 33) {
		print STDERR "\tResetting CBT since got an error related to CBT.\n";
		return $reset_cbt;
	} else {
		my $backup_size_in_mb = $disks_backup_size/1024/1024;
		print STDERR "Current Backup size in MBs is - $backup_size_in_mb\n";
		$xmlroot->appendTextChild('backupSizeMb', $backup_size_in_mb);

		$xmlroot->addChild($disks_xmlelem);
		print STDERR "Returning all disks\n";
		return @disks;
	}

}

sub process_changed_disk_areas {
	my $xmldoc = shift;
	my $xmlroot = shift;
	my $xmlelem = shift;
	my $key = shift;
	my $result = shift;

	my $size;
	my %result_hash = %$result;

	if (defined($result_hash{'changedArea'})) {
		# changed blocks were found
		my $num_areas = $result_hash{'changedArea'};  # arrayref
		my @deref_areas = @$num_areas;

		my $area_count = scalar @deref_areas;
		$xmlelem->appendTextChild('num_areas', $area_count);

		foreach (@deref_areas) {
			my $xmlblockelem = $xmldoc->createElement('changed_disk_area');
			$xmlblockelem->appendTextChild('length', $_->{'length'});
			$xmlblockelem->appendTextChild('start', $_->{'start'});
			$xmlelem->addChild($xmlblockelem);
		}
		
		$size = $result_hash{'length'};
	} else {
		# the whole disk is empty (it has never been used)
		$size = 0;
	}

	return $size;

}

sub get_change_id_from_snap ($$) {
	my $snap_moref = shift;
	my $device_key = shift;

	print STDERR "entering get_change_id_from_snap......\n";
	my $snap_view = Vim::get_view(mo_ref => $snap_moref);

	my @devices = $snap_view->config->hardware->device;
	my $i = 0;
	while ($i < scalar $devices[0]) {
		# TODO: examine this further
		if ($devices[0][$i]->key != $device_key) {
			$i = $i + 1;
			next;
		}
		return $devices[0][$i]->backing->changeId;
	}
}


sub get_changed_blocks_for_disk ($$$$$$$$$$$) {
	my ($vm_moref, $snap_moref, $devicekey, $changeid, $is_template, $xmldoc, $xmlroot, $is_full, $promote_next_ref, $capacityInKB, $snapshots_exist) = @_;

	print STDERR "entering get_changed_blocks_for_disk\n";

	my $capacityInBytes = $capacityInKB * 1024;

	print STDERR "Disk capacityInKB is $capacityInKB\n";
	print STDERR "Disk capacityInBytes is $capacityInBytes\n";
	print STDERR "Snapshots Exist? ===> $snapshots_exist.\n";

	my $disk_backup_size = 0;

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	# Call QCDA only for a VM Backup; not for a Template Backup
	if ($is_template eq "NO") {
		# Check vm_view->capabilities so we know if we can call queryChangedDiskAreas
		if (!$vm_view->capability->changeTrackingSupported) {
			# Server doesn't know about this capability, so it's not supported.
			$xmlroot->appendTextChild('changeTrackingEnabled', '0');
			return (1,$capacityInBytes);
		}
		if (!$vm_view->config->changeTrackingEnabled) {
			# Server has this capability, but the VM doesn't.
			$xmlroot->appendTextChild('changeTrackingEnabled', '0');
			return (1,$capacityInBytes);
		}

		my $nfs_stop = 0;

		# Is deviceKey an RDM disk?
		# If so, skip QCDA, backup whole disk
		my $devices = $vm_view->config->hardware->device;
		my $devCapacity = 0;

		foreach my $dev (@$devices) {
			if ($dev->key == $devicekey) {
				#Bug 15878. Get disk capacity to be used in getting all change-block-areas.
				$devCapacity = ($dev->capacityInKB * 1024);

				if ($is_full and $dev->backing->isa('VirtualDiskRawDiskMappingVer1BackingInfo')) {
					print STDERR "Found RDM disk $devicekey, skipping QCDA() for this device\n";
					$changeid = get_change_id_from_snap($snap_moref, $devicekey);
					print STDERR "Found RDM changeId " . $changeid . " from snapshot.\n";
					if (defined $changeid) {
						$xmlroot->appendTextChild('changeId', $changeid);
					}
					$xmlroot->appendTextChild('changeTrackingSkipped', '1');
					$xmlroot->appendTextChild('changeTrackingEnabled', '1');   # Skip QCDA() later.
					return (1,$devCapacity);
				}

				# Is this device on an NFS datastore?
				if (device_is_on_nfs($dev) and $is_full) {
					# Don't call QueryChangedDiskAreas if the disk is on NFS and we are doing a full.
					# Only supported for diff/incr backups.

					print STDERR "Skipping QueryChangedDiskAreas() since device " . $dev->key . " is on NFS, and this is a full backup.\n";
					# Report change tracking as enabled, but don't use it now
					# This is so the changeId gets recorded.
					$xmlroot->appendTextChild('changeTrackingEnabled', '1');

					# Record that we aren't using change tracking for this particular backup.
					$xmlroot->appendTextChild('changeTrackingSkipped', '1');

					# Will backup the complete disk
					$disk_backup_size = $devCapacity;

					$nfs_stop = 1;
				} else {
					$xmlroot->appendTextChild('changeTrackingSkipped', '0');
				}

				last;  # Only one device will match anyway, stop looking
			}
		}

		my $change_id = get_change_id_from_snap($snap_moref, $devicekey);
		if (defined $change_id) {
			$xmlroot->appendTextChild('changeId', $change_id);
		}

		if ($nfs_stop) {
			# Don't actually use QueryChangedDiskAreas
			return (1,$capacityInBytes);
		}

		if (!defined $changeid and $is_full) {
			$changeid = "*";
		}

		if (!defined $changeid and !$is_full) {
			# This used to set changeid = "*", now we want to fail the backup, since
			# this only happens if a disk is queried which we do not have a change id for.
			$$promote_next_ref = 1;
			die "Disk $devicekey was added to this VM -- a full backup is required.\n";  # TODO message
		}

		VProtect::log_msg("Calling QueryChangedDiskAreas with snapshot: " . $snap_moref->value . ", deviceKey: " . $devicekey . ", changeId: " . $changeid);

		my $startOffset = 0;
		my $areaSize = 0;
		my $processChanges = 1;

		eval {
			my @result;
			my $xmlElem;
			#Bug 16009. Only add changed_disk_areas element if there changes to backup
			#for the current disk. 'addElem' determines if the element is to be added.
			my $addElem = 0;

			#Bug 15878. New loop.
			#Call QueryChangedDiskAreas until we span the entire disk capacity.
			#May require more than 1 call to gather all change-block-areas.
			while  ($processChanges == 1) {
				@result = $vm_view->QueryChangedDiskAreas(
					snapshot => $snap_moref,
					deviceKey => $devicekey,
					startOffset => $startOffset,
					changeId => $changeid);

				if ($startOffset == 0) {
					$xmlElem = $xmldoc->createElement('changed_disk_areas');
				}

				$areaSize = process_changed_disk_areas($xmldoc, $xmlroot, $xmlElem, $devicekey, @result);
				print STDERR "QueryChangedDiskAreas returns start " . $startOffset . " : Size " . $areaSize . "\n";

				if ($areaSize > 0) {
					$startOffset += $areaSize;
					$addElem = 1;
				} else {
					$processChanges = 0;
				}

				if ($startOffset >= $devCapacity) {
					$processChanges = 0;
				}	
			}

			if ($addElem == 1) {
				use bigint;

                                my @starts = ();
                                my @lengths = ();

				foreach my $elem ($xmlElem->findnodes('changed_disk_area')) {
                                    print STDERR "Original start: " . $elem->findvalue('start') . "\tlength: " . $elem->findvalue('length') . "\n";
                                    my $start_value = $elem->findvalue('start');
                                    my $lower = ($start_value / 1048576)*1048576;
					#We need to find the actual position, and figure out the next boundary
					#on that position, then calculate the new length, we just can't
					#arbitrarily extend it out to the next one meg boundary.
					#find the length in the physical file.
                                    my $length_value = $elem->findvalue('length') + $start_value;
                                    my $upper = $length_value;
					#Now calculate to the next 1 meg boundary for the physical
					#locate in the file.
                                    if (($length_value % 1048576) gt 0) {
                                        $upper = (1048576-($length_value % 1048576))+$length_value;
                                    }
					#now calculate the actual true length we need.
				    $upper = $upper - $lower;
                                    print STDERR "Modified start: " . $lower . "\tlength: " . $upper . "\n";
				
				    if ($lower < $capacityInBytes) {
                                    	push(@starts, $lower);
					if (($lower + $upper) > $capacityInBytes) {
						# This means that we are trying to access something out of bounds
						$upper = $capacityInBytes - $lower;
						print STDERR "Modified start: " . $lower . "\tlength: " . $upper . " (last)\n";
					}
                                    	push(@lengths, $upper);
				    }
                                }

                                print STDERR "Cnt0 = " . scalar(@starts) . "\n";
                                
                                for (my $i=0; $i < scalar(@starts); $i++) {
                                    print STDERR "WP step1 start: " . $starts[$i] . "\tlength: " . $lengths[$i] . "\n";
                                }

                                
                                my $prevStart = 0;
                                my $prevLength = 0;
                                my $currStart = 0;
                                my $currLength = 0;
                                
                                my @newStarts = ();
                                my @newLengths = ();

                                for (my $i=0; $i < scalar(@starts); $i++) {
                                    if ($i==0) {
                                        $prevStart = $starts[$i];
                                        $prevLength = $lengths[$i];
                                        next;
                                    }
                                    $currStart = $starts[$i];
                                    $currLength = $lengths[$i];
                                    if ($prevStart != $currStart) {
                                        push(@newStarts, $prevStart);
                                        push(@newLengths, $prevLength);
                                        $prevStart = $currStart;
                                        $prevLength = $currLength;
                                    } else {
                                        if ($prevLength < $currLength) {
                                            $prevLength = $currLength;
                                        }
                                    }
                                }
                                push(@newStarts, $prevStart);  # To take care of the last one
                                push(@newLengths, $prevLength);
                                
                                print STDERR "Cnt1 = " . scalar(@newStarts) . "\n";
                                for (my $i=0; $i < scalar(@newStarts); $i++) {
                                    print STDERR "WP step2 start: " . $newStarts[$i] . "\tlength: " . $newLengths[$i] . "\n";
                                }


                                my $revised_xmlElem = $xmldoc->createElement('changed_disk_areas');
                                $revised_xmlElem->appendTextChild('num_areas', scalar(@newStarts));

                                for (my $i=0; $i < scalar(@newStarts); $i++) {
                                    my $xmlblockelem = $xmldoc->createElement('changed_disk_area');
                                    $xmlblockelem->appendTextChild('start', $newStarts[$i]);
                                    $xmlblockelem->appendTextChild('length',$newLengths[$i]);
                                    $revised_xmlElem->addChild($xmlblockelem);
					$disk_backup_size = $disk_backup_size + $newLengths[$i];
                                }
                                $xmlroot->addChild($revised_xmlElem);
                            }

			$xmlroot->appendTextChild('changeTrackingEnabled', '1');
                    };
		if ($@) {
			# QueryChangedDiskAreas (QCDA) returned a failure

			if (ref ($@) eq 'SoapFault') {
				my $params = "snapshot: " . Dumper($snap_moref) .
					" deviceKey: " . $devicekey .
					" startOffset: 0 " .
					" changeId : " . $changeid;

				print STDERR "Fault type: " . ref ($@) . "\n";
				print STDERR "SoapFault type: " . ref($@->detail) . "\n";

				if (ref($@->detail) eq 'FileFault') {
					print STDERR "FileFault:\n" . Dumper($@) . "\n";
					print STDERR "Params: \n" . $params . "\n";

					if ( ($changeid eq "*") and (!$snapshots_exist) ) {
						#die "Error getting changed disk areas, FileFault.\n\n" .
						#$@ . "\n" . $params . "\n";

						# Error getting CBT info on a FULL Backup of a VM with no snapshots
						# Reset the CBT flag on this VM on the fly and perform
						# the Backup using the new CBT information

						# Return a special error code
						print STDERR "\t\tReturning SPECIAL error code.\n";
						return (33, 0); # TODO Change this if required

					} else {
						$$promote_next_ref = 1;

						die "Error getting changed disk areas: FileFault.\n" . 
						"This issue may be resolved by removing all snapshots and " .
						"performing a Full backup of this VM.\n\n" .
						"PASSTHROUGHPlease see http://support.unitrends.com/ikm/questions.php?questionid=1438 " .
						"for possible steps to resolve this problem.\n\n" .
						$@ . "\n" . $params . "\n";
					}

				} elsif (ref($@->detail) eq 'NoPermission') {
					print STDERR "NoPermissionFault:\n" . Dumper($@) . "\n";
					print STDERR "Params: \n" . $params . "\n";
					die "Error getting changed disk areas: NoPermissionFault.\n" .
						"This issue may be resolved by adding role permissions for " .
						$@->detail->privilegeId . " in your vCenter server.\n\n";
				}
			}
			die (Dumper($@)) . "\n";
		}
	} # End 'if is_template' loop
	else {
		print STDERR "Backup everything for a Template since can't snapshot it\n";
		$xmlroot->appendTextChild('changeTrackingEnabled', '0');
		$xmlroot->appendTextChild('changeTrackingSkipped', '1');
	}

	return (1, $disk_backup_size); # Success
}

sub split_change_ids ($) {
	my $change_ids = shift;

	if (!defined $change_ids) {
		return;
	}

	my @fields = split(/:/, $change_ids);

	my %hash;
	my $hashref = \%hash;

	while (scalar @fields > 0) {
		my $a = shift @fields;
		my $b = shift @fields;

		$hash{$a} = $b;
	}

	return $hashref;
}

sub split_keys ($) {
	my $keys = shift;

	my @array1;
	my $arrayref = \@array1;

	if (!defined $keys) {
		return $arrayref;
	}

	my @fields = split(/:/, $keys);

	my $i = 0;
	while (scalar @fields > 0) {
		$array1[$i] = shift @fields;
		$i = $i + 1;
	}

	return $arrayref;
}

sub guest_vss_operation {
	my $command = shift;
	my $vmName = shift;
	my $vmxPath = shift;
	my $server = shift;
	my $hostUser = shift;
	my $hostPasswd = shift;
	my $vmUser = shift;
	my $vmPasswd = shift;
	my $backupStatus = shift;
	my $warning_ref = shift;
	my $hostPath;
	my $hostName;
	my $hostPort;
	my $err;
	my $cmdArgs;
	my $snapstatus;
	my %procinfo;
	#my $toolsConfPath = "ProgramData\\VMware\\VMware tools\\tools.conf";


	($err, $gos_hostHandle) = HostConnect(
				VIX_API_VERSION,
				VIX_SERVICEPROVIDER_VMWARE_VI_SERVER,
				$server,
				$hostPort, 
				$hostUser, 
				$hostPasswd,
				0,
				VIX_INVALID_HANDLE);

	print STDERR "HostConnect: " . $server . " : " . $hostUser . ":  returns " . $err . " \n";
	if ($err != 0) {
		my $warnMsg;
		$$warning_ref .= "\n Failed to connect to host for guest vss operations. Incorrect credentials. " . GetErrorText($err);
		return $err;
	}	

	($err, $gos_vmHandle) = VMOpen($gos_hostHandle, $vmxPath);
	print STDERR "VMOpen of " . $vmName . " returns " . $err . " \n";
	if ($err != 0) {
		$$warning_ref .= "\n Failed to connect with guest " . $vmName . " for guest vss operations. " . GetErrorText($err); 
		HostDisconnect($gos_hostHandle);
		return $err;
	}

	$err = VMLoginInGuest($gos_vmHandle, $vmUser, $vmPasswd, 0);
	print STDERR "VMLogin returns " . $err . " \n";
	if ($err != 0) {
		$$warning_ref .= "\n Failed to login to guest " . $vmName . " for guest vss operations. " . GetErrorText($err);
		HostDisconnect($gos_hostHandle);
		return $err;
	}

	my $winDir;
	my $arch;
	my $exists;

	## Get some important env variables from guest.
	($err, $winDir) = VMReadVariable($gos_vmHandle, VIX_GUEST_ENVIRONMENT_VARIABLE, "WINDIR", 0);
	print STDERR "ReadVarible returns : err " . $err . " : var : " . $winDir . " \n";
	if ($winDir eq "") {
		## Super.  Cannot read the %windir% env variable from the guest.
		## To find the Windows directory, scan thru the drive letters until
		## we find it.
		my $vl;
		for $vl ("A".."Z") {
			$winDir = $vl . ":\\Windows";
			($err, $exists) = VMDirectoryExistsInGuest($gos_vmHandle, $winDir);
			if ($exists) {
				print STDERR "WINDIR = " . $winDir . "\n";
				last;
			} else {
				$winDir = "";
			}
		}
	}

	## No Windows directory? No problem. Warn the backup an continue without performing the VSS snap.
	if ($winDir eq "") {
		$$warning_ref = $$warning_ref . "\n Failed to find Windows directory in guest. " . $err;
		VMLogoutFromGuest($gos_vmHandle);
		HostDisconnect($gos_hostHandle);
		return 1;
	}

	my $winvolIdx = index($winDir, '\\');
	my $winVol = substr($winDir,0,$winvolIdx+1);

	## Determine the processor architecture so we can run the correct VSS program there.
	($err, $arch) = VMReadVariable($gos_vmHandle, VIX_GUEST_ENVIRONMENT_VARIABLE, "PROCESSOR_ARCHITECTURE", 0);
	print STDERR "ReadVarible returns : err " . $err . " : var : " . $arch . " \n";
	if ($arch eq "") {
		my $wowdir;
		$wowdir = $winDir . "\\SysWOW64";
		($err, $exists) = VMDirectoryExistsInGuest($gos_vmHandle, $wowdir);
		if ($exists) {
			print STDERR "Windows\\SysWOW64 EXISTS\n";
			$arch = "AMD64";
		} else {
			print STDERR "Windows\\SysWOW64 does not EXISTS\n";
			$arch = "X86";
		}
	}

	## Place the VSS agent in %WinDir%\Windows\Unitrends
	$gos_guestPath = $winDir . "\\Unitrends";
	$gos_guestVssPgmPath = $gos_guestPath . "\\VProtectVss.exe";

	($err, $exists) = VMDirectoryExistsInGuest($gos_vmHandle, $gos_guestPath);
	if (!$exists) {
		## Make the Unitrends directory
		my $mkDirCmd = $winDir . "\\system32\\cmd.exe ";
		$cmdArgs = " /C mkdir " . $gos_guestPath;
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
					$mkDirCmd,
					$cmdArgs,
					0,
					VIX_INVALID_HANDLE);	
		print STDERR "guest mkdir of : " . $gos_guestPath . " returns " . $err . "  \n";
		if ($err != 0) {
			$$warning_ref = $$warning_ref . "\n Failed to create temporary Unitrends directory in guest. " . GetErrorText($err);
			VMLogoutFromGuest($gos_vmHandle);
			HostDisconnect($gos_hostHandle);
			return $err;
		}
	}

	## Copy the GetOSver program over and execute to get OS version.
	$gos_osLevel = 6;
	$err = VMCopyFileFromHostToGuest($gos_vmHandle, "/usr/bp/bin/vprotect/GetOsVer.bat", $gos_guestPath . "\\GetOsVer.bat", 0, VIX_INVALID_HANDLE);
	print STDERR "VMCopyFile of : GetOsVer.bat returns " . $err . " \n";
	if ($err == 0) {
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
				$winDir . "\\system32\\cmd.exe ",
				"/C " . $gos_guestPath . "\\GetOsVer.bat",
				0,
				VIX_INVALID_HANDLE);	
		$gos_osLevel = $procinfo{'EXIT_CODE'};
		print STDERR "GetOSVersion returns : " . $gos_osLevel . " \n";
	}

	if ($command eq "create") {
		## Creating the snapshot

		#if Win 2003 or earlier always run the 32-bit binaries.
		if ($gos_osLevel == 5) {
			$arch = "X86";
		}
		if ($arch eq "AMD64") {
			$hostPath = "/usr/bp/bin/vprotect/VProtectVss64.exe";
		} else {
			$hostPath = "/usr/bp/bin/vprotect/VProtectVss.exe";
		}

		#if ($gos_osLevel >= 6) {
			## Insert vmware tools config file that disables vmware vss during our backup.
			## Applies to Windows vista/2008 and later.
		#	$err = VMCopyFileFromHostToGuest($gos_vmHandle, "/usr/bp/bin/vprotect/tools.conf", $winVol . $toolsConfPath, 0, VIX_INVALID_HANDLE);
		#	print STDERR "VMCopyFile " . $toolsConfPath . " returns " . $err . " \n";
		#} 
			
		## Copy the VSS program over.
		$err = VMCopyFileFromHostToGuest($gos_vmHandle, $hostPath, $gos_guestVssPgmPath, 0, VIX_INVALID_HANDLE);
		print STDERR "VMCopyFile of : " . $hostPath . " returns " . $err . " \n";
		if ($err != 0) {
			$$warning_ref .= "\n Failed to copy vss snapshot server to guest. Cannot continue with snapshot. " . GetErrorText($err);
			VMLogoutFromGuest($gos_vmHandle);
			HostDisconnect($gos_hostHandle);
			return $err;
		}

		## Run the VSS program in server mode first. Leaving it running awaiting commands
		## to create vss snapshots and then later remove them and exit.
		$cmdArgs = " Server  " . $gos_guestPath . "\\VssLogSVR.log";
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
					$gos_guestVssPgmPath,
					$cmdArgs,
					VIX_RUNPROGRAM_RETURN_IMMEDIATELY,
					VIX_INVALID_HANDLE);	
		print STDERR "RunPgmInGuest " . $gos_guestVssPgmPath . " " . $cmdArgs . " returns " . $err . " \n";
	
		if ($err != 0)  {
			$$warning_ref .= sprintf("\n Failed to run vss snapshot server in guest. Cannot create VSS snapshot. " . GetErrorText($err)); 
		} else {
			$cmdArgs = " CreateSnap " . $gos_guestPath . "\\VssLogClnt.log";
			($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
						$gos_guestVssPgmPath,
						$cmdArgs,
						0,
						VIX_INVALID_HANDLE);	
			$snapstatus = $procinfo{'EXIT_CODE'}; 
			if (($err != 0) or ($snapstatus != 0))  {
				## Display Windows hex error code.
				if ($snapstatus == 1) {
					$$warning_ref .= "\n Failed to run vss snapshot create program in guest. Guest VSS snapshots are not supported for guest OS.\n";
				} else {	
					$$warning_ref .= sprintf("\n Failed to run vss snapshot create program in guest. Error 0x%x, VSS program returns error 0x%x\n", $err & 0xFFFFFFFF,  $snapstatus & 0xFFFFFFFF);
				}
			}
		}

		print STDERR sprintf("RunPgmInGuest %s %s returns 0x%x : error code 0x%x\n",$gos_guestVssPgmPath,$cmdArgs,$err & 0xFFFFFFFF,$snapstatus & 0xFFFFFFFF);

		## Failure to create snapshot is fatal.
		if (($err != 0) or ($snapstatus != 0)) {
			VMLogoutFromGuest($gos_vmHandle);
			HostDisconnect($gos_hostHandle);
			if ($snapstatus != 0) {
				$err = $snapstatus;
			}
			return $err;
		}
	} else {
		## Removing the snapshot
		my %procinfo;
		my $hostLogFile;
		my $guestLogFile;
		$snapstatus = 0;

		if ($backupStatus eq "SUCCESS") {
			$cmdArgs = " ReleaseOk " . $gos_guestPath . "\\VssLogClnt.log";
		} else {
			$cmdArgs = " ReleaseFail " . $gos_guestPath . "\\VssLogClnt.log";
		}
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
					$gos_guestVssPgmPath,
					$cmdArgs,
					0,
					VIX_INVALID_HANDLE);	
		$snapstatus = $procinfo{'EXIT_CODE'}; 

		if (($err != 0) or ($snapstatus != 0)) {
			$$warning_ref .= sprintf("\n Failed to run vss snapshot release in guest. Error 0x%x : exit code 0x%x\n", $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);
		} 
		print STDERR sprintf("RunPgmInGuest %s %s returns 0x%x : exit code 0x%x\n", $gos_guestVssPgmPath, $cmdArgs, $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);

		$cmdArgs = " Exit " . $gos_guestPath . "\\VssLogClnt.log";
		($err, %procinfo) = VMRunProgramInGuestEx($gos_vmHandle, 
					$gos_guestVssPgmPath,
					$cmdArgs,
					0,
					VIX_INVALID_HANDLE);	
		print STDERR "RunPgmInGuest (shutdown vss agent) " . $gos_guestVssPgmPath . " " . $cmdArgs . " returns " . $err . " \n";
		$snapstatus = $procinfo{'EXIT_CODE'}; 

		if (($err != 0) or ($snapstatus != 0)) {
			$$warning_ref .= sprintf("\n Failed to run vss snapshot exit in guest. Error 0x%x : exit code 0x%x\n", $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);
		} 
		
		print STDERR sprintf("RunPgmInGuest %s %s returns 0x%x : exit code 0x%x\n", $gos_guestVssPgmPath, $cmdArgs, $err & 0xFFFFFFFF, $snapstatus & 0xFFFFFFFF);

		$vmName =~ s/[^a-zA-Z0-9-.#!@%_]*//g;
		$guestLogFile = $gos_guestPath . "\\VssLogSVR.log";
		$hostLogFile = "/usr/bp/logs.dir/vprotect_" . $vmName . ".vssSvr.log";
		$err = VMCopyFileFromGuestToHost($gos_vmHandle, $guestLogFile, $hostLogFile, 0, VIX_INVALID_HANDLE);
		print STDERR "VMCopyFile of : " . $guestLogFile . " returns " . $err . " \n";

		$guestLogFile = $gos_guestPath . "\\VssLogClnt.log";
		$hostLogFile = "/usr/bp/logs.dir/vprotect_" . $vmName . ".vssClnt.log";
		$err = VMCopyFileFromGuestToHost($gos_vmHandle, $guestLogFile, $hostLogFile, 0, VIX_INVALID_HANDLE);
		print STDERR "VMCopyFile of : " . $guestLogFile . " returns " . $err . " \n";

		#if ($gos_osLevel >= 6) {
		#	$err = VMDeleteFileInGuest($gos_vmHandle, $winVol . $toolsConfPath);
		#	print STDERR "remove of file " . $winVol . $toolsConfPath . " : returns " . $err . "\n";
		#}

		## Remove our working directory and all contents on the guest.
		$err = VMDeleteDirectoryInGuest($gos_vmHandle, $gos_guestPath,0);
		print STDERR "Deletion of working directory " . $gos_guestPath . " status : " . $err . "\n";
	}

	VMLogoutFromGuest($gos_vmHandle);
	HostDisconnect($gos_hostHandle);

	return $snapstatus;	
}

# kas 9/16.2012
# Added.
sub do_guest_vss_snapshot {
	my $vmuuid = shift;
	my $server = shift;
	my $hostUser = shift;
	my $hostPasswd = shift;
	my $vmUser = shift;
	my $vmPasswd = shift;
	my $backupStatus = shift;
	my $operation = shift;
	my $warning_ref = shift;

	my $vm;
	my $guestObj;
	my $vmxPath;
	my $err;
	my $pwrState;
	my $vmName;
	my $warning = "";

	print STDERR "Entering create_guest_vss_snapshot\n";

	my $vmmoref = VProtect::vim_get_vm_by_uuid($vmuuid);
	$vm = Vim::get_view(mo_ref => $vmmoref);

	if (not defined ($vm->config)) {
		$$warning_ref = $$warning_ref . "\n Unable to retrieve VM configuration. (Config not defined.)\n";
		return 1;
	}

	$vmName = $vm->name;

	$pwrState = $vm->runtime->powerState->val;
	# Fetch path to vmx file for later use by guest vss operations
	$vmxPath = $vm->config->files->vmPathName;
	print STDERR "VM " . $vmName . " : power state is " . $pwrState . " \n";

	if ($pwrState eq 'poweredOn') {
		print STDERR "VM " . $vmName . " powered on: path: " . $vmxPath . " \n";
		$err = guest_vss_operation($operation, $vmName, $vmxPath, $server, $hostUser, $hostPasswd, $vmUser, $vmPasswd, $backupStatus, \$warning);
		if ($err != 0) {
			$$warning_ref .= $warning;
		}
	} else {
		printf STDERR "VM " . $vmName . " not powered - skipping guest VSS snapshot.\n";
		$err = 0;
	}

	return $err;
}

sub do_backup {
	my $backuptype = shift;
	my $vmuuid = shift;
	my $change_ids = shift;
	my $prev_hw_hash = shift;
	my $serverAddr = shift;
	my $appAwareFlag = shift;
	my $hostUserName = shift;
	my $hostPassword = shift;
	my $gosUserName = shift;
	my $gosPassword = shift;
	my $keys = shift;	# Exclude Keys
	my $is_template = shift;
	my $xmldoc = shift;
	my $xmlroot = shift;

	my $warning = "";

	my $vm_moref;
	my $snap_moref;
	my $err;
	my $retry = 0;

	print STDERR "Original keys of disks to exclude - " . $keys . "\n";

	if ((defined ($change_ids)) and ($is_template eq "YES")) {
		die "Incremental and Differential Backups are not supported for VMware Templates.\n";
	}

	my $promote_next_backup_because_cbt_failed = 0;
	my $promote_next_backup_because_cbt_off = 0; # Used for Incr/Diff Backups

	my $change_id_hash = split_change_ids($change_ids);

	my $key_array = split_keys($keys);

	if (not defined($key_array)) {
		$key_array = "";
	}

	print STDERR "Got args:  appAware: " . $appAwareFlag . " : gosUser : " . $gosUserName . " \n";
	print STDERR "Backup entity is a Template - $is_template\n";

	if ($is_template eq "NO") {
		# Get a moref for the VM.
		$vm_moref = get_moref_by_uuid($vmuuid, $xmldoc, $xmlroot);
	} else {
		# Get a moref for the Template
		$vm_moref = get_template_moref_by_uuid($vmuuid, $xmldoc, $xmlroot);
	}

	my $new_hw_hash = VProtect::get_vm_information_by_moref($vm_moref, $key_array, $xmldoc, $xmlroot);

	my $is_full = 1;
	if (defined $change_ids) {
		$is_full = 0;
	}

	validate_hw_hash($prev_hw_hash, $new_hw_hash, $is_full, $xmldoc, $xmlroot, $vm_moref);

	check_for_fault_tolerance($vm_moref);

	RETRY:
	if ($is_template eq "NO") {
		clean_up_snapshots($vm_moref, $xmldoc, $xmlroot, \$warning);

		configure_change_tracking($vm_moref, $is_full, $xmlroot, \$warning, $retry, \$promote_next_backup_because_cbt_off);
	}

	if ($promote_next_backup_because_cbt_off) {
		VProtect::promote_next_reason($xmldoc, 'CBT found OFF for an Incremental/Differential Backup.');
		die "Change Bloack Tracking (CBT) must be enabled before running a Differential or an Incremental backup.\n" .
		"The next Backup of this Virtual Machine will be promoted to a Full to enable CBT.\n";
	}

	## KAS
	## If guest vss snapshot fails, then update the $warning var to
	## make sure the backup completes minimally with a warning status.
	##
	if ($appAwareFlag eq "YES") {
		if ($is_template eq "YES") {
			die "Cannot do app-aware processing for a Template.\n" .
			"Convert this Template to a Virtual Machine to use app-aware capability.\n";
		} else {
			print STDERR "Doing app-aware guest processing.\n";
			$err = do_guest_vss_snapshot($vmuuid, $serverAddr, $hostUserName, $hostPassword, $gosUserName, $gosPassword, 0, "create", \$warning);
			print STDERR "Value of err is $err\n";
			if ($err != 0) {
				my $remWarn;

				do_guest_vss_snapshot($vmuuid, $serverAddr, $hostUserName, $hostPassword, $gosUserName, $gosPassword, 0, "remove", \$remWarn);
				die $warning . "\n App aware snapshots enabled, but failed to create guest snapshot. Aborting backup.\n";
			}
			print STDERR "Guest vss snapshot warnings:  " . $warning . "\n";
		}
	} else {
		# Set err to a non-zero value for non-appaware Backups, so that create_snapshot sub-routine works in a normal way
		$err = 1;
	}

	my $vm_view = Vim::get_view(mo_ref => $vm_moref);

	my $snapshots_exist = snapshots_exist($vm_view);

	my $has_independent_disks = independent_disks_present($vm_view);

	if ($is_template eq "NO") {
		$snap_moref = create_snapshot($vm_moref, $xmldoc, $xmlroot, \$warning, $err, $has_independent_disks);
	}

	if (!(defined $snap_moref) and ($is_template eq "NO")) {
		print STDERR "create_snapshot returns empty moref\n";
	}

	# Use the default/best transport available transport mode or NBD
	transport_mode($xmlroot);

	my @diskinfo;
	eval {
		@diskinfo = get_disk_info($vm_moref, $snap_moref, $change_id_hash, $key_array, $is_template, $xmldoc, $xmlroot, $is_full, \$promote_next_backup_because_cbt_failed, $snapshots_exist);

		print STDERR "Return value from get_disk_info is -\n" .Dumper(@diskinfo). "\n";
		if ($diskinfo[0] == 33) {
			if ($retry == 0) {
				print STDERR "Removing Unitrends_backup snapshot to retry\n";
				$retry = $retry + 1;
				VProtect::remove_snapshot($snap_moref->{'value'}, $vmuuid);
				goto RETRY;
			} else {
				print STDERR "Not making another attempt to get CBT information.\n";
			}
		}

	};
	if ($@) {
		my @old_error = $@;
		if ($is_template eq "NO") {
			$err = do_guest_vss_snapshot($vmuuid, $serverAddr, $hostUserName, $hostPassword, $gosUserName, $gosPassword, 0, "remove", \$warning);
			# Remove snapshot and throw again
			print STDERR "get_disk_info() failed, removing snapshot with moref " . $snap_moref->{'value'} . "\n";
			VProtect::remove_snapshot($snap_moref->{'value'}, $vmuuid);
			if ($promote_next_backup_because_cbt_failed) {
				VProtect::promote_next_reason($xmldoc, 'A VM changed blocked tracking error has occurred.');
			}
		}
		die (@old_error);
	}

	if ($is_template eq "NO") {
		my $snap_xmlelem = $xmldoc->createElement('snapshot');
		$snap_xmlelem->appendTextChild('snap_moref', $snap_moref->{'value'});
		$snap_xmlelem->appendTextChild('snap_moref_type', $snap_moref->{'type'});
		$xmlroot->addChild($snap_xmlelem);
	}

	if ($warning ne "") {
		# Put any warnings we generated into the XML.
		VProtect::output_xml_warning($xmldoc, $warning);
	}
}

sub do_backup_conclude {
	my $vmuuid = shift;
	my $serverAddr = shift;
	my $appAwareFlag = shift;
	my $hostUserName = shift;
	my $hostPassword = shift;
	my $gosUserName = shift;
	my $gosPassword = shift;
	my $backupStatus = shift;
	my $warning = "";
	my $vm_moref;
	my $snap_moref;
	my $err = 0;

	print STDERR "Got args:  appAware: " . $appAwareFlag . " : gosUser : " . $gosUserName . " \n";

	if ($appAwareFlag eq "YES") {
		# Get a moref for the VM.
		print STDERR "About to call into do_guest_vss_snapshot\n";
		$err = do_guest_vss_snapshot($vmuuid, $serverAddr, $hostUserName, $hostPassword, $gosUserName, $gosPassword, $backupStatus, "release", \$warning);
	}
	
	print STDERR "Leaving do_backup_conclude\n";

	return $err;
}

1;
